/**
 * Author: Dave Toth
 * Date: Fall 2005
 * 
 * This program acts as a validation tool for a stripped-down public-resource
 * computing client.  The validator consists of several threads. The main (control)
 * thread stops and starts the other threads based on what they would be doing at
 * a given time.  These times are read in from the trace files.
 *
 * The volunteer computing client portion of the program requests work from 
 * the server and receives a file name as a response.  The client then 
 * downloads the file indicated by the response from the web server over 
 * http, saves the file to the local disk, opens the file from the
 * local disk, reads the contents, and pocesses the file using a set of 
 * computations.  Finally, the client returns an answer to the server.
 */

import java.net.*;      // Use sockets.
import java.io.*;       // Read and write over sockets.
import java.util.*;     // Timers, Vectors, Hashtable
import java.util.Timer; // Timers
import java.awt.*;      // Timers


public class ValidatorClient {

    //////////////////////////////////////////////////////////////////////
    //                            Constants                             //
    //////////////////////////////////////////////////////////////////////

    /**
     * Indicates that the file was not downloaded successfully.
     */
    public static final int DOWNLOAD_FAILURE = -999;

    /**
     * Indicates that the computation failed.
     */
    public static final int FAILURE = -1;

    /**
     * Indicates that the computation succeeded.
     */
    public static final int SUCCESS = 1;

    /**
     * Indicates the failure of an operation that returns a boolean.
     */
    public static final boolean FAILED = false;

    /**
     * Indicates the success of an operation that returns a boolean.
     */
    public static final boolean OK = true;

    /**
     * Indicates that the client is requesting work.
     */
    public static final String REQUEST_WORK = "WORK_BEING_REQUESTED";

    /**
     * Indicates that the client is returning the result of a computation.
     */
    public static final String RETURN_RESULT = "RESULT_BEING_RETURNED";

    /**
     * Indicates that the client has successfully downloaded the file.
     */

    public static final String ACKNOWLEDGE_DOWNLOAD = "DOWNLOAD_COMPLETE";
    /**
     * Indicates that it's time to shut the server down cleanly so we 
     * don't trash any data by breaking database connections in an 
     * unclean way.
     */
    public static final String SHUTDOWN_SERVER = "SHUTDOWN_REQUESTED";

    /**
     * The value inidcating that an array slot will not be used.
     */
    public static final int UNINITIALIZED = -1;

    /**
     * The constants representing the state of the computer that's simulating the volunteer computing client.
     */
    public static final int OFF = 0;
    public static final int ON_UN = 1;
    public static final int AVAIL = 2;

    /**
     * The constants representing the task retrieval methods.
     */
    public static final int BUFFER_NONE = 0;
    public static final int DOWNLOAD_EARLY = 1;
    public static final int BUFFER_1_TASK = 2;
    public static final int BUFFER_1_DAY = 3;
    public static final int BUFFER_3_DAYS = 4;
    public static final int BUFFER_7_DAYS = 5;
    public static final int BUFFER_14_DAYS = 6;

    private static long timeSimulationStarted;

    //////////////////////////////////////////////////////////////////////
    //                          Globals                                 //
    //////////////////////////////////////////////////////////////////////


    /**
     * The Timer we use to control the threads.
     */
    Timer myTimer;

    /**
     * The index we use to point to the appropriate array slot so we set the Timer to the correct value.
     */
    int timerIndex;

    /**
     * One fewer than the number of periods that the computer had during the trace time.
     */
    int periods;

    /**
     * The durations of the periods in the traces.  There's only 765 lines
     * in the biggest trace, so we can use 767 slots and be covered.
     */
    int durations[];

    /**
     * The states of the periods in the traces.  There's only 765 lines
     * in the biggest trace, so we can use 767 slots and be covered.
     */
    int states[];

    /**
     * The thread for the science application.
     */
    ScienceApplication scienceApplication;

    /**
     * The delay bound.
     */
    long delayBound;

    /**
     * The task retrieval method.
     */
    int taskRetrievalMethod;

    /**
     * The trace file to use.
     */
    static String traceFile;

    /**
     * The state of the client (ON-UN, OFF, AVAIL) during the current period.
     */
    int currentState;
    
    /**
     * The state of the client (ON-UN, OFF, AVAIL) during the next period.
     */
    int nextState;
    
    /**
     * The vector where we hold references to the threads.
     */ 
    static Vector threadVector;

    /**
     * The vector that holds the tasks that need to be completed.
     */
    static Vector taskVector;

    /*
     * The unique ID of the client.
     */
    String clientID;

    /**
     * This Hashtable holds the percent of time the cpu is available for 
     * volunteer computing for each of the traces.
     */
    public static Hashtable cpuAvailabilityTable;

    /**
     * The number of tasks completed by the volunteer computing client.
     */
    private static int tasksCompleted = 0;

    /**
     * A reference to the ValidatorClient.
     */
    private ValidatorClient validatorClient;



    //////////////////////////////////////////////////////////////////////
    //                          Methods                                 //
    //////////////////////////////////////////////////////////////////////
 

    /**
     * The main method.  Creates an instance of a prc client and hands
     * control over to the client.
     * @param args - The array of command line arguments passed to the JVM.
     *               To automate testing, we set the first argument as the
     *               file size and the second argument as the computation size.
     *               
     *               The arguments are: 
     *                 1) The DelayBound.
     *                 2) The Task Retrieval Method.
     *                 3) The trace file.
     */
    public static void main(String[] args) {
	try {
	    Thread.sleep(1000*300*Integer.parseInt(args[3]));
	}
	catch (InterruptedException ie) {
	    System.out.println("Initial sleep for unsynching clients interrupted");
	}

	ValidatorClient client = new ValidatorClient(Long.parseLong(args[0]), Integer.parseInt(args[1]), args[2]);
    }


    /**
     * Constructor.  Create the client and runs it.
     */
    public ValidatorClient(long delayBound, int taskRetrievalMethod, String traceFile) {

	////////////////////////////////////////////////////////////////////
	//                        INITIALIZATION                          //
	////////////////////////////////////////////////////////////////////

	validatorClient = this;

	this.delayBound = delayBound;
	this.taskRetrievalMethod = taskRetrievalMethod;
	this.traceFile = traceFile;

	threadVector = new Vector();
	taskVector = new Vector();

	System.out.println("Starting validator");



	cpuAvailabilityTable = new Hashtable();
	// Debugging
	cpuAvailabilityTable.put("work3.txt", new Float(72));
	cpuAvailabilityTable.put("student_99.trc", new Float(100));
	// End Debugging

	cpuAvailabilityTable.put("business_trace_1.trc", new Float(1.96));
	cpuAvailabilityTable.put("business_trace_10.trc", new Float(75.97));
	cpuAvailabilityTable.put("business_trace_11.trc", new Float(4.02));
	cpuAvailabilityTable.put("business_trace_12.trc", new Float(8.77));
	cpuAvailabilityTable.put("business_trace_13.trc", new Float(10.43));
	cpuAvailabilityTable.put("business_trace_14.trc", new Float(5.39));
	cpuAvailabilityTable.put("business_trace_15.trc", new Float(4.06));
	cpuAvailabilityTable.put("business_trace_16.trc", new Float(53.22));
	cpuAvailabilityTable.put("business_trace_17.trc", new Float(51.68));
	cpuAvailabilityTable.put("business_trace_18.trc", new Float(17.97));
	cpuAvailabilityTable.put("business_trace_19.trc", new Float(70.51));
	cpuAvailabilityTable.put("business_trace_2.trc", new Float(76.96));
	cpuAvailabilityTable.put("business_trace_20.trc", new Float(.67));
	cpuAvailabilityTable.put("business_trace_21.trc", new Float(81.06));
	cpuAvailabilityTable.put("business_trace_22.trc", new Float(51.98));
	cpuAvailabilityTable.put("business_trace_23.trc", new Float(8.63));
	cpuAvailabilityTable.put("business_trace_24.trc", new Float(21.22));
	cpuAvailabilityTable.put("business_trace_25.trc", new Float(3.39));
	cpuAvailabilityTable.put("business_trace_26.trc", new Float(60.69));
	cpuAvailabilityTable.put("business_trace_3.trc", new Float(7.31));
	cpuAvailabilityTable.put("business_trace_4.trc", new Float(9.36));
	cpuAvailabilityTable.put("business_trace_5.trc", new Float(37.46));
	cpuAvailabilityTable.put("business_trace_6.trc", new Float(8.26));
	cpuAvailabilityTable.put("business_trace_7.trc", new Float(3.33));
	cpuAvailabilityTable.put("business_trace_8.trc", new Float(11.15));
	cpuAvailabilityTable.put("business_trace_9.trc", new Float(44.21));
	cpuAvailabilityTable.put("ccc_trace_1.trc", new Float(70.12));
	cpuAvailabilityTable.put("ccc_trace_10.trc", new Float(58.78));
	cpuAvailabilityTable.put("ccc_trace_11.trc", new Float(64.08));
	cpuAvailabilityTable.put("ccc_trace_12.trc", new Float(61.81));
	cpuAvailabilityTable.put("ccc_trace_13.trc", new Float(66.11));
	cpuAvailabilityTable.put("ccc_trace_14.trc", new Float(61.23));
	cpuAvailabilityTable.put("ccc_trace_15.trc", new Float(65.65));
	cpuAvailabilityTable.put("ccc_trace_16.trc", new Float(66.32));
	cpuAvailabilityTable.put("ccc_trace_17.trc", new Float(63.13));
	cpuAvailabilityTable.put("ccc_trace_18.trc", new Float(62.56));
	cpuAvailabilityTable.put("ccc_trace_19.trc", new Float(67.45));
	cpuAvailabilityTable.put("ccc_trace_2.trc", new Float(68.17));
	cpuAvailabilityTable.put("ccc_trace_20.trc", new Float(65.69));
	cpuAvailabilityTable.put("ccc_trace_21.trc", new Float(73.20));
	cpuAvailabilityTable.put("ccc_trace_22.trc", new Float(74.31));
	cpuAvailabilityTable.put("ccc_trace_23.trc", new Float(77.92));
	cpuAvailabilityTable.put("ccc_trace_24.trc", new Float(71.06));
	cpuAvailabilityTable.put("ccc_trace_25.trc", new Float(71.14));
	cpuAvailabilityTable.put("ccc_trace_26.trc", new Float(71.96));
	cpuAvailabilityTable.put("ccc_trace_27.trc", new Float(84.29));
	cpuAvailabilityTable.put("ccc_trace_28.trc", new Float(75.99));
	cpuAvailabilityTable.put("ccc_trace_29.trc", new Float(66.95));
	cpuAvailabilityTable.put("ccc_trace_3.trc", new Float(65.59));
	cpuAvailabilityTable.put("ccc_trace_30.trc", new Float(68.15));
	cpuAvailabilityTable.put("ccc_trace_31.trc", new Float(75.90));
	cpuAvailabilityTable.put("ccc_trace_32.trc", new Float(81.54));
	cpuAvailabilityTable.put("ccc_trace_33.trc", new Float(80.75));
	cpuAvailabilityTable.put("ccc_trace_34.trc", new Float(65.05));
	cpuAvailabilityTable.put("ccc_trace_35.trc", new Float(69.43));
	cpuAvailabilityTable.put("ccc_trace_36.trc", new Float(67.18));
	cpuAvailabilityTable.put("ccc_trace_37.trc", new Float(70.40));
	cpuAvailabilityTable.put("ccc_trace_38.trc", new Float(62.69));
	cpuAvailabilityTable.put("ccc_trace_39.trc", new Float(68.29));
	cpuAvailabilityTable.put("ccc_trace_4.trc", new Float(63.83));
	cpuAvailabilityTable.put("ccc_trace_40.trc", new Float(67.77));
	cpuAvailabilityTable.put("ccc_trace_41.trc", new Float(75.62));
	cpuAvailabilityTable.put("ccc_trace_42.trc", new Float(63.15));
	cpuAvailabilityTable.put("ccc_trace_43.trc", new Float(89.18));
	cpuAvailabilityTable.put("ccc_trace_44.trc", new Float(79.19));
	cpuAvailabilityTable.put("ccc_trace_45.trc", new Float(91.91));
	cpuAvailabilityTable.put("ccc_trace_46.trc", new Float(82.68));
	cpuAvailabilityTable.put("ccc_trace_47.trc", new Float(84.43));
	cpuAvailabilityTable.put("ccc_trace_48.trc", new Float(78.81));
	cpuAvailabilityTable.put("ccc_trace_49.trc", new Float(82.42));
	cpuAvailabilityTable.put("ccc_trace_5.trc", new Float(66.52));
	cpuAvailabilityTable.put("ccc_trace_50.trc", new Float(88.03));
	cpuAvailabilityTable.put("ccc_trace_51.trc", new Float(82.53));
	cpuAvailabilityTable.put("ccc_trace_52.trc", new Float(88.97));
	cpuAvailabilityTable.put("ccc_trace_53.trc", new Float(83.86));
	cpuAvailabilityTable.put("ccc_trace_54.trc", new Float(79.55));
	cpuAvailabilityTable.put("ccc_trace_55.trc", new Float(75.12));
	cpuAvailabilityTable.put("ccc_trace_56.trc", new Float(88.15));
	cpuAvailabilityTable.put("ccc_trace_57.trc", new Float(73.08));
	cpuAvailabilityTable.put("ccc_trace_58.trc", new Float(91.54));
	cpuAvailabilityTable.put("ccc_trace_59.trc", new Float(89.41));
	cpuAvailabilityTable.put("ccc_trace_6.trc", new Float(73.31));
	cpuAvailabilityTable.put("ccc_trace_60.trc", new Float(81.64));
	cpuAvailabilityTable.put("ccc_trace_61.trc", new Float(82.98));
	cpuAvailabilityTable.put("ccc_trace_62.trc", new Float(85.34));
	cpuAvailabilityTable.put("ccc_trace_63.trc", new Float(84.29));
	cpuAvailabilityTable.put("ccc_trace_64.trc", new Float(80.12));
	cpuAvailabilityTable.put("ccc_trace_65.trc", new Float(80.58));
	cpuAvailabilityTable.put("ccc_trace_66.trc", new Float(81.99));
	cpuAvailabilityTable.put("ccc_trace_67.trc", new Float(92.53));
	cpuAvailabilityTable.put("ccc_trace_68.trc", new Float(84.04));
	cpuAvailabilityTable.put("ccc_trace_7.trc", new Float(58.40));
	cpuAvailabilityTable.put("ccc_trace_8.trc", new Float(61.91));
	cpuAvailabilityTable.put("ccc_trace_9.trc", new Float(62.12));
	cpuAvailabilityTable.put("home_trace_1.trc", new Float(36.68));
	cpuAvailabilityTable.put("home_trace_10.trc", new Float(69.85));
	cpuAvailabilityTable.put("home_trace_12.trc", new Float(.15));
	cpuAvailabilityTable.put("home_trace_13.trc", new Float(48.21));
	cpuAvailabilityTable.put("home_trace_14.trc", new Float(50.32));
	cpuAvailabilityTable.put("home_trace_15.trc", new Float(1.18));
	cpuAvailabilityTable.put("home_trace_16.trc", new Float(36.16));
	cpuAvailabilityTable.put("home_trace_17.trc", new Float(.69));
	cpuAvailabilityTable.put("home_trace_18.trc", new Float(1.04));
	cpuAvailabilityTable.put("home_trace_2.trc", new Float(2.91));
	cpuAvailabilityTable.put("home_trace_21.trc", new Float(36.88));
	cpuAvailabilityTable.put("home_trace_22.trc", new Float(2.95));
	cpuAvailabilityTable.put("home_trace_23.trc", new Float(45.51));
	cpuAvailabilityTable.put("home_trace_24.trc", new Float(38.33));
	cpuAvailabilityTable.put("home_trace_25.trc", new Float(87.12));
	cpuAvailabilityTable.put("home_trace_3.trc", new Float(17.48));
	cpuAvailabilityTable.put("home_trace_4.trc", new Float(75.59));
	cpuAvailabilityTable.put("home_trace_5.trc", new Float(75.28));
	cpuAvailabilityTable.put("home_trace_6.trc", new Float(37.18));
	cpuAvailabilityTable.put("home_trace_7.trc", new Float(10.05));
	cpuAvailabilityTable.put("home_trace_9.trc", new Float(4.96));
	cpuAvailabilityTable.put("student_10.trc", new Float(5.29));
	cpuAvailabilityTable.put("student_11.trc", new Float(13.53));
	cpuAvailabilityTable.put("student_12.trc", new Float(49.99));
	cpuAvailabilityTable.put("student_14.trc", new Float(80.95));
	cpuAvailabilityTable.put("student_15.trc", new Float(59.43));
	cpuAvailabilityTable.put("student_16.trc", new Float(36.25));
	cpuAvailabilityTable.put("student_17.trc", new Float(85.38));
	cpuAvailabilityTable.put("student_19.trc", new Float(51.19));
	cpuAvailabilityTable.put("student_2.trc", new Float(3.82));
	cpuAvailabilityTable.put("student_20.trc", new Float(62.82));
	cpuAvailabilityTable.put("student_21.trc", new Float(.05));
	cpuAvailabilityTable.put("student_22.trc", new Float(32.90));
	cpuAvailabilityTable.put("student_23.trc", new Float(7.21));
	cpuAvailabilityTable.put("student_26.trc", new Float(.03));
	cpuAvailabilityTable.put("student_27.trc", new Float(8.41));
	cpuAvailabilityTable.put("student_30.trc", new Float(7.74));
	cpuAvailabilityTable.put("student_31.trc", new Float(59.43));
	cpuAvailabilityTable.put("student_32.trc", new Float(6.93));
	cpuAvailabilityTable.put("student_33.trc", new Float(.61));
	cpuAvailabilityTable.put("student_34.trc", new Float(2.65));
	cpuAvailabilityTable.put("student_36.trc", new Float(13.88));
	cpuAvailabilityTable.put("student_38.trc", new Float(1.47));
	cpuAvailabilityTable.put("student_4.trc", new Float(47.24));
	cpuAvailabilityTable.put("student_5.trc", new Float(68.55));
	cpuAvailabilityTable.put("student_7.trc", new Float(12.19));


	// Read in Times from Trace file.

	// Initialize the states and durations arrays.  We set everything to
	// UNINITIALIZED and then as we read in the values, we place them in the 
	// arrays.  This will let us tell when there are no more array slots that
	// will be used.
	durations = new int[766];
	states = new int[766];

	for (int i = 0; i < 766; i ++) {
	    states[i] = UNINITIALIZED;
	    durations[i] = UNINITIALIZED;
	}

	periods = -1;

	// Open the file and read in the values form the trace.
	try { 
	    BufferedReader input = new BufferedReader(new FileReader(traceFile));
	    String line;
	    int duration;
	    while ((line = input.readLine()) != null) {
		if (line.indexOf("ON-UN") >= 0) { // ON-UN
		    periods ++;
		    duration = Integer.parseInt(line.substring(1 + line.indexOf(",")));
		    states[periods] = ON_UN;
		    durations[periods] = duration;
		}
		else if (line.indexOf("OFF") >= 0) { // OFF
		    periods ++;
		    duration = Integer.parseInt(line.substring(1 + line.indexOf(",")));
		    states[periods] = OFF;
		    durations[periods] = duration;
		}
		else if (line.indexOf("AVAIL") >= 0) { // AVAIL
		    periods ++;
		    duration = Integer.parseInt(line.substring(1 + line.indexOf(",")));
		    states[periods] = AVAIL;
		    durations[periods] = duration;
		}
		else if (line.indexOf("Trace Time") >= 0) { // Trace Time
		    // DO NOTHING
		}
		else { // Blank Line
		    // DO NOTHING
		}
	    }
	}
	catch (Exception e) {
	    System.out.println("Failure reading trace file.  Exiting.");
	    System.exit(0);
	}


	// Get the clientID
	clientID = "";

	try {
	    InetAddress localIP = InetAddress.getLocalHost();
	    clientID = (localIP.getHostAddress()).substring((localIP.getHostAddress()).lastIndexOf(".") + 1);
	}
	catch (UnknownHostException uhe) {
	    System.out.println("Unknow Host.");
	    System.exit(0);
	}

	////////////////////////////////////////////////////////////////////
	//                      END INITIALIZATION                        //
	////////////////////////////////////////////////////////////////////


	timeSimulationStarted = System.currentTimeMillis();

	// IF THE FIRST STATE IS ON, CREATE THE SCIENCE APPLICATION AND START IT
	// OTHERWISE, JUST START THE TIMER
	currentState = states[0];
	if (currentState == AVAIL) {
	    scienceApplication = new ScienceApplication(taskRetrievalMethod, clientID, delayBound, validatorClient);
	    scienceApplication.threadPaused = true;
	    threadVector.add(scienceApplication);
	    scienceApplication.start();
	}

	// Control threads as trace data gets used.
	// Once we kick off this Timer, it will transfer control of the program over to 
	// the inner classes below that update the Threads.  When there are no more 
	// periods to do, the program will exit.
	myTimer = new Timer();
	timerIndex = 0;
	myTimer.schedule(task, durations[timerIndex]*1000);
	timerIndex ++;
    }

    public void incrementTasksCompleted() {
	tasksCompleted ++;
    }

    public int getTasksCompleted() {
	return tasksCompleted;
    }

    public static long getCurrentSimulationTime() {
	long tempTime = System.currentTimeMillis();
	return ((tempTime - timeSimulationStarted)/1000);
    }



    ////////////////////////////////////////////////////////////////////
    // INNER CLASSES
    ////////////////////////////////////////////////////////////////////



    /**
     * The TimerTask is the class that controls the program flow.
     * Every time one completes, it calls the updateThreads class
     * below which starts a new TimerTask if there are more 
     * periods left to run.  
     */
    TimerTask task = new TimerTask() {
	    public void run() {
		EventQueue.invokeLater(updateThreads);
	    }
	};
    
    
    /**
     * This object creates a new TimerTask if there are more periods to run.
     * Otherwise, it forces the program to exit.
     */
    final Runnable updateThreads = new Runnable() {
	    public void run() {
		task = null;
		
		TimerTask task = new TimerTask() {
			public void run() {
			    EventQueue.invokeLater(updateThreads);
			}
		    };
		if (timerIndex <= periods) {
		    // Update the state machine that keeps track of what should happen to the threads and then make it so.
		    nextState = states[timerIndex];
		    if ((currentState == AVAIL) && (nextState == ON_UN)) {
			//System.out.println("AVAIL->ON-UN: " + getCurrentSimulationTime());
			// PAUSE THREADS
			currentState = nextState;
			for (int j = 0; j < threadVector.size(); j ++) {
			    if (threadVector.elementAt(j) instanceof ScienceApplication) {
				ScienceApplication th = (ScienceApplication) threadVector.elementAt(j);
				th.threadPaused = true;		
				th.interrupt();
			    }
			    else {
				if (threadVector.elementAt(j) instanceof DownloadThread) {
				    DownloadThread th = (DownloadThread) threadVector.elementAt(j);
				    //th.threadPaused = true;
				    //th.interrupt();			
				}	
			    }
			}
		    }
		    else if ((currentState == AVAIL) && (nextState == OFF)) {
			//System.out.println("AVAIL->OFF: " + getCurrentSimulationTime());
			// HALT THREADS & DESTROY THEM
			currentState = nextState;
			for (int j = 0; j < threadVector.size(); j ++) {
			    if (threadVector.elementAt(j) instanceof ScienceApplication) {
				ScienceApplication th = (ScienceApplication) threadVector.elementAt(j);
				th.threadPaused = false;
				th.interrupt();
			    }
			    else {
				if (threadVector.elementAt(j) instanceof DownloadThread) {
				    DownloadThread th = (DownloadThread) threadVector.elementAt(j);
				    th.threadPaused = false;
				    th.interrupt();				
				}
			    }
			}
			threadVector.clear();
		    }
		    else if ((currentState == ON_UN) && (nextState == AVAIL)) {
			//System.out.println("ON-UN->AVAIL: " + getCurrentSimulationTime());
			// IF THREADS EXIST RESTART THEM, OTHERWISE CREATE THEM AND START THEM
			currentState = nextState;
			if (threadVector.size() > 0) {
			    // Threads exist so restart them.
			    for (int j = 0; j < threadVector.size(); j ++) {
				if (threadVector.elementAt(j) instanceof ScienceApplication) {
				    ScienceApplication th = (ScienceApplication) threadVector.elementAt(j);
				    th.threadPaused = true;
				    th.interrupt();
				}
				else {
				    if (threadVector.elementAt(j) instanceof DownloadThread) {
					DownloadThread th = (DownloadThread) threadVector.elementAt(j);
					th.threadPaused = true;
					th.interrupt();				
				    }			
				}
			    }
			}
			else {
			    // Threads don't exist so create them.
			    scienceApplication = new ScienceApplication(taskRetrievalMethod, clientID, delayBound, validatorClient);
			    scienceApplication.threadPaused = true;
			    threadVector.add(scienceApplication);
			    scienceApplication.start();
			}
		    }
		    else if ((currentState == ON_UN) && (nextState == OFF)) {
			//System.out.println("ON-UN->OFF: " + getCurrentSimulationTime());
			// IF THREADS EXIST STOP THEM
			currentState = nextState;
			for (int j = 0; j < threadVector.size(); j ++) {
			    if (threadVector.elementAt(j) instanceof ScienceApplication) {
				ScienceApplication th = (ScienceApplication) threadVector.elementAt(j);
				th.threadPaused = false;
				th.interrupt();
			    }
			    else {
				if (threadVector.elementAt(j) instanceof DownloadThread) {
				    DownloadThread th = (DownloadThread) threadVector.elementAt(j);
				    th.threadPaused = false;
				    th.interrupt();			
				}	
			    }
			}
			threadVector.clear();
		    }
		    else if ((currentState == OFF) && (nextState == AVAIL)) {
			//System.out.println("OFF->AVAIL: " + getCurrentSimulationTime());
			// CREATE THREADS AND START THEM
			currentState = nextState;
			scienceApplication = new ScienceApplication(taskRetrievalMethod, clientID, delayBound, validatorClient);
			scienceApplication.threadPaused = true;
			threadVector.add(scienceApplication);
			scienceApplication.start();
		    }
		    else if ((currentState == OFF) && (nextState == ON_UN)) {
			//System.out.println("OFF->ON-UN: " + getCurrentSimulationTime());
			// Do Nothing.
			currentState = nextState;
		    }
		    else { // The current state and next state are the same.  Don't do anything.
			// Do nothing.
			currentState = nextState;
		    }

		    myTimer.schedule(task, durations[timerIndex]*1000);
		    timerIndex ++;
		}
		else {
		    System.out.println("DONE");
		    System.out.println("FFTs Completed: " + tasksCompleted);
		    System.out.println("Portion of Most Recent Task: " + scienceApplication.fftCompleted + "      as a percent: " + ((((double) scienceApplication.fftCompleted)/((double) scienceApplication.FFTS_IN_WORKUNIT))*((double) 100)));
		    System.exit(0);
		}
	    }
	};
}
